#!/usr/bin/env python3

"""
soc_helper
Copyright (C) 2023  M. Williges (spam at zut punkt de)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import paho.mqtt.client as mqtt
import paho.mqtt.publish as publish
import json
import time
import datetime
import configuration
from os.path import exists

#if configuration.USE_SPRITMONITOR:
#    import spritmonitor

#Klasse, um eine zusammengehörige Datengruppe zu speichern
class ladevorgangclass:
    def __init__(self):
        self.erase()
        
    def erase(self):
        self.counter = -1
        self.counterAtPlugIn = -1
        self.counterAtPlugOut = -1
        self.gelanden = 0
        self.odo = 0
        self.trip = 0
        self.soc = 0
        self.openwbsoc = 0
        self.plug = 0
        self.datetimeAtPlugOut = 0

    def setCounter(self, e):
        self.counter = e
    
    def setOdo(self, o):
        self.odo = o
    
    def setSoc(self, s):
        self.soc = s
    
    def setOpenwbsoc(self, s):
        self.openwbsoc = s
    
    def setPlug(self, s):
        #A change in plug state may indicate the begin or end of a charge
        if (self.plug == 0) and (s != 0):
            #Stecker wurde eingesteckt. Energiezähler sichern
            self.counterAtPlugIn = self.counter
            output(2,'Ladestecker wurde gesteckt bei Zähler '+str(self.counterAtPlugIn))
        elif (self.plug !=0) and (s==0):
            #Stecker wurde abgesteckt.
            self.counterAtPlugOut = self.counter	#merken des Zählerstandes beim aktuellen Ende der Ladung
            output(2,'Ladestecker wurde entfernt bei Zähler '+str(self.counterAtPlugOut))
            # Auswertung des Ladevorgangs.
            self.datetimeAtPlugOut = datetime.datetime.now()	#Datum des Ladeendes
            self.geladen = round( (self.counterAtPlugOut - self.counterAtPlugIn)/configuration.OPENWB_FACTORCOUNTER, 3)	#Zähler auf kWh umrechnen und mit 3 Nachkommastellen runden
            output(1,'Es wurden '+str(self.geladen)+'kWh geladen.')
            try:
                self.save()
            except:
                output(0,'lokales Abspeichern fehlgeschlagen!')
            
        elif (self.plug != 0) and (self.counterAtPlugIn<0) and (self.counter>=0):
            # der Stecker steckte schon, als das Programm gestartet wurde und es liegt noch kein Zählerstand bei Einstecken vor.
            # In diesem Fall wird der aktuelle Zählerstand gespeichert, falls er mindestens einmal übertragen wurde.
            output(2,'Stecker steckt, es liegt aber noch kein Zählerwert vor.')
            self.counterAtPlugIn = self.counter
            output(2,'Zählerstand bei Anstecken wird auf aktuellen Zählerstand gesetzt: '+str(self.counter))
            
        #Zustand des Steckers sichern
        self.plug = s
        
    def save(self):
        # Speichern des Ladevorgangs in der Logdatei
        global savefile
        savefile.write(str(self.datetimeAtPlugOut.strftime('%d.%m.%Y'))+', '+str(self.odo)+', '+str(self.geladen)+', '+str(self.soc)+', '+(self.openwbsoc.decode())+'\n')
        savefile.flush()

#Ausgabefunktion. Stellt die aktuelle Systemzeit vor die Ausgabe und veranlasst die sofortige Ausgabe, statt zu buffern
def output(loglevel, text):
    if loglevel <= configuration.LOGLEVEL:
        dt=datetime.datetime.now()
        print(dt,text,flush=True)

#Callback-funktion, die bei erfolgreicher Verbindung zur OpenWB aufgerufen wird (CONNACK)
def on_connect(client, userdata, flags, rc):
    output(2,"Connected with result code "+str(rc))
    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    client.subscribe('#')

# Funktion, die permanent den Energiezähler in eine interne Variable überträgt. Die Übertragung erfolgt in der Regel so schnell,
# daß dem Empfang von Kilometerstand / SOC immer ein aktueller Zählerstand zugeordnet werden kann.
def cb_energycounter(client, userdata, msg):
    ladevorgang.setCounter(float(msg.payload))

# Beobachtung des Steckerzustandes des Ladekabels. Ein Ladevorgang gilt als beendet, wenn der Stecker von gesteckt zu nicht gesteckt wechselt
def cb_plug(client, userdata, msg):
    if (b'true' in msg.payload) or (b'1' in msg.payload):
        ladevorgang.setPlug(1)
        # Ausgabe unterdrückt, da Zustand alle 1,75s übertragen wird
        #output(2,'Stecker steckt')
    else:
        ladevorgang.setPlug(0)
        # Ausgabe unterdrückt, da Zustand alle 1,75s übertragen wird
        #output(2,'Stecker steckt nicht')

# Auslesen des SOC aus der Wallbox - dient der Beurteilung, ob nach Ladeende eine Vollbetankung vorliegt
def cb_openwbsoc(client, userdata, msg):
    ladevorgang.setOpenwbsoc(msg.payload)
        
# Callback-Funktion für WiCAN-Statusbotschaften. Wenn ein status:online vom wican empfangen wird, wird
# unmittelbar danach eine SOC- und eine ODO-Abfrage geschickt.
def cb_status(client, userdata, msg):
    output(2,'Status: '+msg.topic+" "+str(msg.payload))
    #Statusbotschaft empfangen. Prüfen, ob online
    try:
        s=json.loads(msg.payload)
        if s['status'] == 'online':
            #Wenn das Fahrzeug online geht, kann der SOC und der Gesamtkm-Stand abgefragt werden
            output(2,'Fahrzeug ist online. Sende SOC- und DST-Anforderung')
            client.publish(configuration.TOPIC_PREFIX+configuration.TOPIC_TX, configuration.SOC_REQUEST)
            client.publish(configuration.TOPIC_PREFIX+configuration.TOPIC_TX, configuration.DST_REQUEST)
        else:
            output(2,'Fahrzeugstatus ist nicht <<online>>')
    except:
        output(1,'Nachricht vom WiCAN enthält kein <status>')

# Callback-Funktion für eine CAN-Rx-Botschaft. Prüft, ob es eine SOC- oder ODO-haltige Botschaft ist
# und sendet den extrahierten Wert zur OpenWB
def cb_rx(client, userdata, msg):
    #CAN-Botschaft empfangen. Prüfen, ob SOC und dann auswerten und and openWB pushen
    output(2,'Empfangene CAN-Botschaft: '+str(msg.payload))
    try:
        d = json.loads(msg.payload)    #string der payload als dict in d
        f = d['frame']  #Datenframe in f ergibt eine Liste mit einem Element
        dd = f[0]    #das nullte und einzige Element der Liste ist ein dict und wird in dd kopiert
        ddd=dd['data']  #das Element data aus dem frame-dict enthält die Nutzbytes vom CAN als list
        if dd['id'] == configuration.VALID_SOC_ID:
            if ddd[1:1+len(configuration.VALID_SOC_ANSWER)] == configuration.VALID_SOC_ANSWER:
                #Umrechnungsformel. Hier bei Bedarf anpassen (das Motorsteuergerät des up meldet unter 0x22 0x11 0x16 mit ID 0x7e8 auch 2 Bytes in der Form (ddd[4]*256+ddd[5])/100
                soc=round(ddd[4]/2.55,1)
                output(1,'Fahrzeug-SOC ist '+str(soc))
                ladevorgang.setSoc(soc)
                output(2,"SOC-Wert von "+str(soc)+' an '+configuration.OPENWB_SETSOC_TOPIC+' schicken.')
                #SOC-Wert an die OpenWB schicken
                publish.single(configuration.OPENWB_SETSOC_TOPIC, soc, hostname=configuration.OPENWB_IP)
            else:
                output(1,'Botschaft ist keine gültige Antwort auf SOC-Anfrage')
        elif dd['id'] == configuration.VALID_DST_ID:
            if ddd[1:1+len(configuration.VALID_DST_ANSWER)] == configuration.VALID_DST_ANSWER:
                #Umrechnungsformel. Hier bei Bedarf anpassen
                odo=ddd[4]*2560+ddd[5]*10
                ladevorgang.setOdo(odo)
                output(1,'Fahrzeug-Kilometerstand ist '+str(odo))
            else:
                output(1,'Kilometerstand-Botschaft ist keine gültige Antwort auf Kilometerstand-Anfrage')
        else:
            output(1,'Empfangene Botschaft trägt keinen bekannten Identifier. Wird ignoriert.')
    except:
        output(1,'Fehler beim Parsen einer CAN-Botschaft vom WiCAN. Wird ignoriert')

#Globale Klasse für Ladedaten erstellen
ladevorgang = ladevorgangclass()

#Datei zum Abspeichern der Ladedaten öffnen
try:
    if not exists(configuration.CHARGELOG_PATH):
        output(2, configuration.CHARGELOG_PATH+' existiert nicht. Lege neu an.')
        savefile = open(configuration.CHARGELOG_PATH,'a')
        savefile.write('Datum, Kilometerstand, Energiemenge, Start-SOC (Auto), End-SOC (Wallbox)\n')
        savefile.flush()
    else:
        output(2, 'Öffne existierende Datei '+configuration.CHARGELOG_PATH)
        savefile = open(configuration.CHARGELOG_PATH,'a')
except:
    output(1,'FATAL: Konnte Ausgabedatei '+configuration.CHARGELOG_PATH+' nicht anlegen oder öffnen')
    quit()

# MQTT-Client einrichten und mit dem Broker verbinden
client = mqtt.Client()
client.on_pre_connect = None
client.on_connect = on_connect
# Abos einrichten und mit Callback-Funktionen verbinden, die aufgerufen werden, wenn ein abonniertes Topic auftaucht.
client.message_callback_add(configuration.OPENWB_CHARGECOUNTER_TOPIC, cb_energycounter)
client.message_callback_add(configuration.OPENWB_GETPLUGSTATE_TOPIC, cb_plug)
client.message_callback_add(configuration.OPENWB_GETSOC_TOPIC, cb_openwbsoc)
client.message_callback_add(configuration.TOPIC_PREFIX+configuration.TOPIC_STATUS+'/#', cb_status)
client.message_callback_add(configuration.TOPIC_PREFIX+configuration.TOPIC_RX+'/#', cb_rx)

try:
    client.connect(configuration.OPENWB_IP, 1883, 60)
except:
    output(1,"Fehler beim Verbinden zum Broker "+configuration.OPENWB_IP)

#Hauptschleife. Es kann entweder eine Endlosschleife genutzt werden oder alle 5 Minuten aktiv der SOC abgefragt werden, sofern das Fahrzeug beim 
#Laden komplett aktiv ist (OBD-Buchse bestromt, Niedervoltbatterie geladen). Den nicht genutzten Code bitte auskommentieren


# Hintergrundschleife, wenn beim Laden das Fahrzeug nicht periodisch abgefragt wird
client.loop_forever()

# alternativ dazu eine Hintergrundschleife, wenn periodisch der SOC abgefragt werden soll.
# Wenn sie genutzt werden soll, bitte die Zeile oben auskommentieren und die Kommentarzeichen vor 
# den folgenden 4 Zeilen entfernen:
#client.loop_start()
#while 0<1:
#    time.sleep(60)
#    client.publish(configuration.TOPIC_PREFIX+configuration.TOPIC_TX, configuration.SOC_REQUEST)
